use cgmath::Point2;
use galactica_content::{Content, FactionHandle, OutfitHandle, ShipHandle, SystemHandle};
use galactica_galaxy::{ship::ShipPersonality, Galaxy, GxShipHandle};
use galactica_playeragent::PlayerAgent;
use galactica_systemsim::{ParticleBuilder, StepResources, SystemSim, Wrapper};
use galactica_util::timing::Timing;
use rand::seq::SliceRandom;
use std::time::Instant;

#[derive(Clone)]
pub struct GameState {
	pub gx: Galaxy,
	pub systemsim: SystemSim,
	pub timing: Timing,
	pub start_instant: Instant,
}

unsafe impl Send for GameState {}
unsafe impl Sync for GameState {}

pub struct Game {
	// Core game data
	ct: Content,
	state: GameState,

	// Metadata
	wrapper: Wrapper, // Physics computer
	time_scale: f32,
	last_update: Instant,

	/// Particles to create this frame.
	/// Must be cleared at the start of every frame
	/// TODO: better way to handle this?
	new_particles: Vec<ParticleBuilder>,
}

unsafe impl<'a> Send for Game {}

impl<'a> Game {
	pub fn make_player(&mut self) -> GxShipHandle {
		let player = self.state.gx.create_ship(
			&self.ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 0 },
			ShipPersonality::Player,
			&SystemHandle { index: 0 },
		);
		let s = self.state.gx.get_ship_mut(player).unwrap();
		s.add_outfit(&self.ct.get_outfit(OutfitHandle { index: 0 }));
		s.add_outfit(&self.ct.get_outfit(OutfitHandle { index: 1 }));
		s.add_outfit(&self.ct.get_outfit(OutfitHandle { index: 2 }));

		self.state
			.systemsim
			.add_ship(&self.ct, &s, Point2 { x: 0.0, y: 0.0 });

		return player;
	}

	pub fn new(ct: Content) -> Self {
		let mut gx = Galaxy::new(&ct);

		let a = gx.create_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 1 },
			ShipPersonality::Dummy,
			&SystemHandle { index: 0 },
		);
		let s = gx.get_ship_mut(a).unwrap();
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 0 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let a = gx.create_ship(
			&ct,
			ShipHandle { index: 0 },
			FactionHandle { index: 0 },
			ShipPersonality::Point,
			&SystemHandle { index: 0 },
		);
		let s = gx.get_ship_mut(a).unwrap();
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 0 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 1 }));
		s.add_outfit(&ct.get_outfit(OutfitHandle { index: 2 }));

		let systemsim = SystemSim::new(&ct, &gx, SystemHandle { index: 0 });

		let state = GameState {
			gx,
			systemsim,
			timing: Timing::new(),
			start_instant: Instant::now(),
		};

		Game {
			ct,
			state,

			wrapper: Wrapper::new(),
			last_update: Instant::now(),
			time_scale: 1.0,
			new_particles: Vec::new(),
		}
	}

	pub fn update_player_controls(&mut self, player: &PlayerAgent) {
		self.state.systemsim.update_player_controls(player)
	}

	pub fn get_state(&self) -> &GameState {
		&self.state
	}

	pub fn get_particles(&self) -> &Vec<ParticleBuilder> {
		&self.new_particles
	}

	pub fn update(&mut self) {
		self.state.timing.start_frame();
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;
		self.new_particles.clear();

		self.state.timing.start_galaxy();
		self.state.gx.step(t);
		self.state.timing.mark_galaxy();

		self.state.systemsim.step(StepResources {
			ct: &self.ct,
			gx: &mut self.state.gx,
			particles: &mut self.new_particles,
			timing: &mut self.state.timing,
			wrapper: &mut self.wrapper,
			t,
		});

		self.last_update = Instant::now();
		self.new_particles.shuffle(&mut rand::thread_rng());
		self.state.timing.mark_frame();
	}
}
