use cgmath::Point2;
use galactica_content::{Content, SystemHandle};
use galactica_galaxy::{Galaxy, GxShipHandle};
use galactica_systemsim::{ParticleBuilder, SystemSim};
use galactica_util::timing::Timing;
use glyphon::{FontSystem, SwashCache, TextAtlas, TextRenderer};
use std::rc::Rc;
use wgpu::BufferAddress;
use winit::window::Window;

use crate::{globaluniform::GlobalUniform, vertexbuffer::VertexBuffer};

/// Bundles parameters passed to a single call to GPUState::render
pub struct RenderInput<'a> {
	/// Camera position, in world units
	pub camera_pos: Point2<f32>,

	/// Player ship data
	pub player_data: GxShipHandle,

	/// The system we're currently in
	pub current_system: SystemHandle,

	/// Height of screen, in world units
	pub camera_zoom: f32,

	/// The world state to render
	pub systemsim: &'a SystemSim,

	// TODO: handle overflow. is it a problem?
	/// The current time, in seconds
	pub current_time: f32,

	/// Game content
	pub ct: &'a Content,

	/// Game data
	pub gx: &'a Galaxy,

	/// Particles to spawn during this frame
	pub particles: &'a Vec<ParticleBuilder>,

	/// Time we spent in each part of the game loop
	pub timing: Timing,
}

/// Renderer state. A reference to this struct is often passed to helper functions.
pub(crate) struct RenderState {
	pub window: Window,
	pub window_size: winit::dpi::PhysicalSize<u32>,
	pub window_aspect: f32,

	pub queue: wgpu::Queue,
	pub global_uniform: GlobalUniform,
	pub vertex_buffers: VertexBuffers,

	pub text_font_system: FontSystem,
	pub text_cache: SwashCache,
	pub text_atlas: TextAtlas,
	pub text_renderer: TextRenderer,
}

/// Vertex buffers
pub(crate) struct VertexBuffers {
	// Keeps track of length of each buffer
	// Most of these are reset on each frame.
	//
	// The exception is particle_counter, which
	// is never reset, and loops back to zero once
	// it exceeds buffer length
	pub object_counter: BufferAddress,
	pub ui_counter: BufferAddress,
	pub particle_counter: BufferAddress,
	pub radialbar_counter: BufferAddress,

	pub object: Rc<VertexBuffer>,
	pub starfield: Rc<VertexBuffer>,
	pub ui: Rc<VertexBuffer>,
	pub particle: Rc<VertexBuffer>,
	pub radialbar: Rc<VertexBuffer>,
}
