//! Various implementations of [`crate::ShipBehavior`]

mod null;
mod point;

use null::*;
use point::PointShipController;

use galactica_galaxy::GxShipHandle;
use rapier2d::dynamics::{RigidBodyHandle, RigidBodySet};
use std::{collections::HashMap, fmt::Debug};

use crate::{
	objects::{ShipControls, SySimShip},
	StepResources,
};

/// Represents a ship controller
#[derive(Debug, Clone)]
pub enum ShipController {
	/// Null controller
	Null(NullShipController),

	/// Point controller
	Point(PointShipController),
}

impl ShipController {
	/// Make a new null controller
	pub fn new_null() -> Self {
		Self::Null(NullShipController::new())
	}

	/// Make a new point controller
	pub fn new_point() -> Self {
		Self::Point(PointShipController::new())
	}

	/// Compute new ship controls from world state
	pub fn update_controls(
		&mut self,
		res: &StepResources,
		rigid_bodies: &RigidBodySet,
		ships: &HashMap<GxShipHandle, SySimShip>,
		this_ship: RigidBodyHandle,
		this_data: GxShipHandle,
	) -> Option<ShipControls> {
		match self {
			Self::Null(n) => n.update_controls(res, rigid_bodies, ships, this_ship, this_data),
			Self::Point(p) => p.update_controls(res, rigid_bodies, ships, this_ship, this_data),
		}
	}
}

/// Ship controller trait. Any struct that implements this
/// may be used to control a ship.
pub trait ShipControllerStruct
where
	Self: Debug + Clone,
{
	/// Update a ship's controls based on system state.
	/// This method returns the ship's new control values,
	/// or None if no change is to be made.
	fn update_controls(
		&mut self,
		res: &StepResources,
		rigid_bodies: &RigidBodySet,
		ships: &HashMap<GxShipHandle, SySimShip>,
		this_ship: RigidBodyHandle,
		this_data: GxShipHandle,
	) -> Option<ShipControls>;
}
