use cgmath::{Deg, InnerSpace};
use galactica_content::Relationship;
use galactica_galaxy::GxShipHandle;
use rapier2d::dynamics::{RigidBodyHandle, RigidBodySet};
use std::collections::HashMap;

use super::ShipControllerStruct;
use crate::{
	objects::{ShipControls, SySimShip},
	util, StepResources,
};

/// "Point" ship controller.
/// Point and shoot towards the nearest enemy.
#[derive(Debug, Clone)]
pub struct PointShipController {}

impl PointShipController {
	/// Create a new ship controller
	pub fn new() -> Self {
		Self {}
	}
}

impl ShipControllerStruct for PointShipController {
	fn update_controls(
		&mut self,
		res: &StepResources,
		rigid_bodies: &RigidBodySet,
		ships: &HashMap<GxShipHandle, SySimShip>,
		this_ship: RigidBodyHandle,
		this_data: GxShipHandle,
	) -> Option<ShipControls> {
		let mut controls = ShipControls::new();

		let this_rigidbody = rigid_bodies.get(this_ship).unwrap();
		let my_data = res.gx.get_ship(this_data).unwrap();
		let my_position = util::rigidbody_position(this_rigidbody);
		let my_rotation = util::rigidbody_rotation(this_rigidbody);
		let my_angvel = this_rigidbody.angvel();
		let my_faction = res.ct.get_faction(my_data.get_faction());

		// Iterate all possible targets
		let mut hostile_ships = ships
			.values()
			.filter(|s| {
				let data = res.gx.get_ship(s.data_handle);
				if let Some(data) = data {
					match my_faction.relationships.get(&data.get_faction()).unwrap() {
						Relationship::Hostile => true,
						_ => false,
					}
				} else {
					// This check is necessary---we don't want to target (or panic on)
					// ships that don't have data (and are thus playing their collapse animation)
					false
				}
			})
			.map(|s| rigid_bodies.get(s.rigid_body).unwrap());

		// Find the closest target
		let mut closest_enemy_position = match hostile_ships.next() {
			Some(c) => util::rigidbody_position(c),
			None => return Some(controls), // Do nothing if no targets are available
		};

		let mut d = (my_position - closest_enemy_position).magnitude();
		for r in hostile_ships {
			let p = util::rigidbody_position(r);
			let new_d = (my_position - p).magnitude();
			if new_d < d {
				d = new_d;
				closest_enemy_position = p;
			}
		}

		let angle_delta: Deg<f32> = (my_rotation)
			.angle(closest_enemy_position - my_position)
			.into();

		if angle_delta < Deg(0.0) && my_angvel > -0.3 {
			controls.right = true;
		} else if angle_delta > Deg(0.0) && my_angvel < 0.3 {
			controls.left = true;
		}

		controls.guns = true;
		return Some(controls);
	}
}
