use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Rad, Vector2, Zero};
use galactica_content::{CollapseEvent, Ship, ShipHandle};
use nalgebra::point;
use rand::{rngs::ThreadRng, Rng};
use rapier2d::{dynamics::RigidBody, geometry::Collider};

use crate::{util, ParticleBuilder, StepResources};

#[derive(Debug, Clone)]
pub(super) struct ShipCollapseSequence {
	total_length: f32,
	time_elapsed: f32,
	rng: ThreadRng,
}

impl ShipCollapseSequence {
	pub(super) fn new(total_length: f32) -> Self {
		Self {
			total_length,
			time_elapsed: 0.0,
			rng: rand::thread_rng(),
		}
	}

	/// Has this sequence been fully played out?
	pub(super) fn is_done(&self) -> bool {
		self.time_elapsed >= self.total_length
	}

	/// Pick a random points inside a ship's collider
	fn random_in_ship(&mut self, ship_content: &Ship, collider: &Collider) -> Vector2<f32> {
		let mut y = 0.0;
		let mut x = 0.0;
		let mut a = false;
		while !a {
			y = self.rng.gen_range(-1.0..=1.0) * ship_content.size / 2.0;
			x = self.rng.gen_range(-1.0..=1.0) * ship_content.size * ship_content.sprite.aspect
				/ 2.0;
			a = collider.shape().contains_local_point(&point![x, y]);
		}
		Vector2 { x, y }
	}

	/// Step this sequence `t` seconds
	pub(super) fn step(
		&mut self,
		res: &mut StepResources,
		ship_handle: ShipHandle,
		rigid_body: &mut RigidBody,
		collider: &mut Collider,
	) {
		let ship_content = res.ct.get_ship(ship_handle);
		let ship_pos = util::rigidbody_position(rigid_body);
		let ship_rot = util::rigidbody_rotation(rigid_body);
		let ship_ang = ship_rot.angle(Vector2 { x: 1.0, y: 0.0 });

		// The fraction of this collapse sequence that has been played
		let frac_done = self.time_elapsed / self.total_length;

		// Trigger collapse events
		for event in &ship_content.collapse.events {
			match event {
				CollapseEvent::Effect(event) => {
					if (event.time > self.time_elapsed && event.time <= self.time_elapsed + res.t)
						|| (event.time == 0.0 && self.time_elapsed == 0.0)
					{
						for spawner in &event.effects {
							let effect = res.ct.get_effect(spawner.effect);

							for _ in 0..spawner.count as usize {
								let pos = if let Some(pos) = spawner.pos {
									pos.to_vec()
								} else {
									self.random_in_ship(ship_content, collider)
								};
								let pos = ship_pos
									+ (Matrix2::from_angle(-ship_ang - Rad::from(Deg(90.0))) * pos);

								let velocity = rigid_body.velocity_at_point(&point![pos.x, pos.y]);

								res.particles.push(ParticleBuilder::from_content(
									effect,
									pos,
									Rad::zero(),
									Vector2 {
										x: velocity.x,
										y: velocity.y,
									},
									Vector2::zero(),
								))
							}
						}
					}
				}
			}
		}

		// Create collapse effects
		for spawner in &ship_content.collapse.effects {
			let effect = res.ct.get_effect(spawner.effect);

			// Probability of adding a particle this frame.
			// The area of this function over [0, 1] should be 1.
			let pdf = |x: f32| {
				let f = 0.2;
				let y = if x < (1.0 - f) {
					let x = x / (1.0 - f);
					(x * x + 0.2) * 1.8 - f
				} else {
					1.0
				};
				return y;
			};

			let p_add = (res.t / self.total_length) * pdf(frac_done) * spawner.count;

			if self.rng.gen_range(0.0..=1.0) <= p_add {
				let pos = if let Some(pos) = spawner.pos {
					pos.to_vec()
				} else {
					self.random_in_ship(ship_content, collider)
				};

				// Position, adjusted for ship rotation
				let pos = ship_pos + Matrix2::from_angle(-ship_ang - Rad::from(Deg(90.0))) * pos;
				let vel = rigid_body.velocity_at_point(&point![pos.x, pos.y]);

				res.particles.push(ParticleBuilder {
					sprite: effect.sprite,
					pos,
					velocity: Vector2 { x: vel.x, y: vel.y },
					angle: Rad::zero(),
					angvel: Rad::zero(),
					lifetime: effect.lifetime,
					size: effect.size,
					fade: 0.0,
				});
			}
		}

		self.time_elapsed += res.t;
	}
}
