use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Rad, Vector2, Zero};
use galactica_content::{Content, FactionHandle};
use galactica_galaxy::GxShipHandle;
use nalgebra::{point, vector};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyHandle},
	geometry::{Collider, ColliderHandle},
};

use crate::{util, ParticleBuilder, StepResources};

use super::collapse::ShipCollapseSequence;

/// A ship's controls
#[derive(Debug, Clone)]
pub struct ShipControls {
	/// True if turning left
	pub left: bool,

	/// True if turning right
	pub right: bool,

	/// True if foward thrust
	pub thrust: bool,

	/// True if firing guns
	pub guns: bool,
}

impl ShipControls {
	/// Create a new, empty ShipControls
	pub fn new() -> Self {
		ShipControls {
			left: false,
			right: false,
			thrust: false,
			guns: false,
		}
	}
}

/// A ship instance in the physics system
#[derive(Debug, Clone)]
pub struct SySimShip {
	/// This ship's physics handle
	pub rigid_body: RigidBodyHandle,

	/// This ship's collider
	pub collider: ColliderHandle,

	/// This ship's game data
	pub data_handle: GxShipHandle,

	/// This ship's controls
	pub(crate) controls: ShipControls,

	/// This ship's collapse sequence
	collapse_sequence: ShipCollapseSequence,

	/// This ship's faction.
	/// This is technically redundant, faction is also stored in
	/// game data, but that's destroyed once the ship dies.
	/// We need the faction for the collapse sequence!
	faction: FactionHandle,
}

impl SySimShip {
	/// Make a new ship
	pub(crate) fn new(
		ct: &Content,
		data_handle: GxShipHandle,
		faction: FactionHandle,
		rigid_body: RigidBodyHandle,
		collider: ColliderHandle,
	) -> Self {
		let ship_content = ct.get_ship(data_handle.content_handle());
		SySimShip {
			rigid_body,
			collider,
			data_handle,
			controls: ShipControls::new(),
			faction,
			collapse_sequence: ShipCollapseSequence::new(ship_content.collapse.length),
		}
	}

	/// If this is true, remove this ship from the physics system.
	pub fn should_be_removed(&self) -> bool {
		self.collapse_sequence.is_done()
	}

	/// Step this ship's state by t seconds
	pub fn step(
		&mut self,
		res: &mut StepResources,
		rigid_body: &mut RigidBody,
		collider: &mut Collider,
	) {
		let ship_data = res.gx.get_ship(self.data_handle);
		if ship_data.is_none() {
			// If ship data is none, it has been removed because the ship has been destroyed.
			// play collapse sequence.
			self.collapse_sequence.step(
				res,
				self.data_handle.content_handle(),
				rigid_body,
				collider,
			);
		} else {
			return self.step_live(res, rigid_body, collider);
		}
	}

	/// Step this ship's state by t seconds (called when alive)
	fn step_live(
		&mut self,
		res: &mut StepResources,
		rigid_body: &mut RigidBody,
		collider: &mut Collider,
	) {
		let ship = res.gx.get_ship(self.data_handle).unwrap();
		let ship_content = res.ct.get_ship(self.data_handle.content_handle());
		let ship_pos = util::rigidbody_position(&rigid_body);
		let ship_rot = util::rigidbody_rotation(rigid_body);
		let ship_ang = ship_rot.angle(Vector2 { x: 1.0, y: 0.0 });
		let mut rng = rand::thread_rng();

		if ship.get_hull() <= ship_content.damage.hull {
			for e in &ship_content.damage.effects {
				if rng.gen_range(0.0..=1.0) <= res.t / e.frequency {
					let effect = res.ct.get_effect(e.effect);

					let pos = if let Some(pos) = e.pos {
						pos.to_vec()
					} else {
						// Pick a random point inside this ship's collider
						let mut y = 0.0;
						let mut x = 0.0;
						let mut a = false;
						while !a {
							y = rng.gen_range(-1.0..=1.0) * ship_content.size / 2.0;
							x = rng.gen_range(-1.0..=1.0)
								* ship_content.size * ship_content.sprite.aspect
								/ 2.0;
							a = collider.shape().contains_local_point(&point![x, y]);
						}
						Vector2 { x, y }
					};

					let pos =
						ship_pos + (Matrix2::from_angle(-ship_ang - Rad::from(Deg(90.0))) * pos);

					let velocity = rigid_body.velocity_at_point(&point![pos.x, pos.y]);

					res.particles.push(ParticleBuilder::from_content(
						effect,
						pos,
						Rad::zero(),
						Vector2 {
							x: velocity.x,
							y: velocity.y,
						},
						Vector2::zero(),
					))
				}
			}
		}

		let engine_force = ship_rot * res.t;

		if self.controls.thrust {
			rigid_body.apply_impulse(
				vector![engine_force.x, engine_force.y] * ship.get_outfits().get_engine_thrust(),
				true,
			);
		}

		if self.controls.right {
			rigid_body
				.apply_torque_impulse(ship.get_outfits().get_steer_power() * -100.0 * res.t, true);
		}

		if self.controls.left {
			rigid_body
				.apply_torque_impulse(ship.get_outfits().get_steer_power() * 100.0 * res.t, true);
		}
	}
}

impl SySimShip {
	/// Get this ship's control state
	pub fn get_controls(&self) -> &ShipControls {
		&self.controls
	}

	/// Get this ship's faction
	pub fn get_faction(&self) -> FactionHandle {
		self.faction
	}
}
