use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Point2, Rad, Vector2, Zero};
use galactica_content::{
	Content, GunPoint, OutfitHandle, ProjectileCollider, Relationship, SystemHandle,
};
use galactica_galaxy::{
	ship::{GxShip, ShipPersonality},
	Galaxy, GxShipHandle,
};
use galactica_playeragent::PlayerAgent;
use nalgebra::{point, vector};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyBuilder, RigidBodyHandle, RigidBodySet},
	geometry::{ColliderBuilder, ColliderHandle, ColliderSet},
	pipeline::ActiveEvents,
};
use std::{collections::HashMap, f32::consts::PI};

use crate::{
	controller::ShipController,
	objects::{SySimProjectile, SySimShip},
	util, ParticleBuilder, StepResources,
};

/// Manages the physics state of one system
#[derive(Clone)]
pub struct SystemSim {
	/// The system this sim is attached to
	_system: SystemHandle,

	rigid_body_set: RigidBodySet,
	collider_set: ColliderSet,

	projectiles: HashMap<ColliderHandle, SySimProjectile>,
	ships: HashMap<GxShipHandle, SySimShip>,

	// TODO: these don't need to be cloned each frame
	ship_behaviors: HashMap<GxShipHandle, ShipController>,
	collider_ship_table: HashMap<ColliderHandle, GxShipHandle>,
}

// Private methods
impl<'a> SystemSim {
	fn remove_projectile(
		&mut self,
		res: &mut StepResources,
		c: ColliderHandle,
	) -> Option<(RigidBody, SySimProjectile)> {
		let p = match self.projectiles.remove(&c) {
			Some(p) => p,
			None => return None,
		};
		let r = self
			.rigid_body_set
			.remove(
				p.rigid_body,
				&mut res.wrapper.im,
				&mut self.collider_set,
				&mut res.wrapper.ij,
				&mut res.wrapper.mj,
				true,
			)
			.unwrap();

		return Some((r, p));
	}

	fn remove_ship(&mut self, res: &mut StepResources, s: ColliderHandle) {
		let s = match self.collider_ship_table.get(&s) {
			None => return,
			Some(s) => match self.ships.get(s) {
				None => return,
				Some(s) => s,
			},
		};

		self.rigid_body_set.remove(
			s.rigid_body,
			&mut res.wrapper.im,
			&mut self.collider_set,
			&mut res.wrapper.ij,
			&mut res.wrapper.mj,
			true,
		);
		let h = self.collider_ship_table.remove(&s.collider).unwrap();
		self.ship_behaviors.remove(&h);
		self.ships.remove(&h);
	}

	fn collide_projectile_ship(
		&mut self,
		res: &mut StepResources,
		projectile_h: ColliderHandle,
		ship_h: ColliderHandle,
	) {
		let projectile = self.projectiles.get(&projectile_h);
		let ship = self
			.ships
			.get_mut(self.collider_ship_table.get(&ship_h).unwrap());
		if projectile.is_none() || ship.is_none() {
			return;
		}
		let projectile = projectile.unwrap();
		let ship = ship.unwrap();

		let f = res.ct.get_faction(projectile.faction);
		let r = f.relationships.get(&ship.get_faction()).unwrap();
		let destory_projectile = match r {
			Relationship::Hostile => {
				// We only apply damage if the target ship is alive
				if let Some(ship_d) = res.gx.get_ship_mut(ship.data_handle) {
					ship_d.apply_damage(projectile.content.damage);
				}
				true
			}
			_ => false,
		};

		if destory_projectile {
			let pr = self.rigid_body_set.get(projectile.rigid_body).unwrap();
			let v = util::rigidbody_velocity(pr).normalize() * projectile.content.force;
			let pos = util::rigidbody_position(pr);
			let _ = pr;

			let r = self.rigid_body_set.get_mut(ship.rigid_body).unwrap();
			r.apply_impulse_at_point(vector![v.x, v.y], point![pos.x, pos.y], true);

			// Borrow again, we can only have one at a time
			let pr = self.rigid_body_set.get(projectile.rigid_body).unwrap();
			let pos = util::rigidbody_position(pr);
			let angle = util::rigidbody_rotation(pr).angle(Vector2 { x: 1.0, y: 0.0 });

			match &projectile.content.impact_effect {
				None => {}
				Some(x) => {
					let effect = res.ct.get_effect(*x);
					let r = ship.rigid_body;
					let sr = self.get_rigid_body(r).unwrap();
					let parent_velocity = util::rigidbody_velocity(pr);
					let target_velocity =
						sr.velocity_at_point(&nalgebra::Point2::new(pos.x, pos.y));

					res.particles.push(ParticleBuilder::from_content(
						effect,
						pos,
						-angle,
						parent_velocity,
						Vector2 {
							x: target_velocity.x,
							y: target_velocity.y,
						},
					));
				}
			};

			self.remove_projectile(res, projectile.collider);
		}
	}
}

// Public methods
impl SystemSim {
	/// Create a new physics system
	pub fn new(ct: &Content, gx: &Galaxy, system: SystemHandle) -> Self {
		let mut w = Self {
			_system: system,
			rigid_body_set: RigidBodySet::new(),
			collider_set: ColliderSet::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
			ship_behaviors: HashMap::new(),
			collider_ship_table: HashMap::new(),
		};

		// TODO: guarantee not touching
		// TODO: add, remove ships each tick
		// Maybe store position in galaxy crate?
		let mut rng = rand::thread_rng();
		for s in gx.iter_ships() {
			w.add_ship(
				ct,
				s,
				Point2 {
					x: rng.gen_range(-500.0..=500.0),
					y: rng.gen_range(-500.0..=500.0),
				},
			);
		}

		return w;
	}

	/// Add a ship to this physics system
	pub fn add_ship(&mut self, ct: &Content, ship: &GxShip, position: Point2<f32>) {
		let ship_content = ct.get_ship(ship.get_content());
		let cl = ColliderBuilder::convex_decomposition(
			&ship_content.collision.points[..],
			&ship_content.collision.indices[..],
		)
		// Rotate collider to match sprite
		// (Collider starts pointing east, sprite starts pointing north.)
		.rotation(PI / -2.0)
		.mass(ship_content.mass);
		// TODO: only build colliders once

		let rb = RigidBodyBuilder::dynamic()
			.angular_damping(ship_content.angular_drag)
			.linear_damping(ship_content.linear_drag)
			.translation(vector![position.x, position.y])
			.can_sleep(false);

		let r = self.rigid_body_set.insert(rb.build());
		let c = self
			.collider_set
			.insert_with_parent(cl.build(), r, &mut self.rigid_body_set);

		self.collider_ship_table.insert(c, ship.get_handle());

		self.ship_behaviors.insert(
			ship.get_handle(),
			match ship.get_personality() {
				ShipPersonality::Dummy | ShipPersonality::Player => ShipController::new_null(),
				ShipPersonality::Point => ShipController::new_point(),
			},
		);

		self.ships.insert(
			ship.get_handle(),
			SySimShip::new(ct, ship.get_handle(), ship.get_faction(), r, c),
		);
	}

	/// Update a player ship's controls
	pub fn update_player_controls(&mut self, player: &PlayerAgent) {
		if player.ship.is_none() {
			return;
		}
		let ship_object = self.ships.get_mut(&player.ship.unwrap());
		if let Some(ship_object) = ship_object {
			ship_object.controls.guns = player.input.key_guns;
			ship_object.controls.left = player.input.key_left;
			ship_object.controls.right = player.input.key_right;
			ship_object.controls.thrust = player.input.key_thrust;
		}
	}

	/// Run ship updates:
	/// - updates ship controls (runs behaviors)
	/// - applies ship controls
	/// - creates projectiles
	fn step_ships(&mut self, res: &mut StepResources) {
		// We can't apply these right away since self is borrowed
		// by the iterator
		// TODO: don't allocate!
		let mut projectiles = Vec::new();
		let mut to_remove = Vec::new();
		for (_, handle) in &self.collider_ship_table {
			// Borrow immutably for now...
			// (required to compute controls)
			let ship_object = self.ships.get(handle).unwrap();
			if ship_object.should_be_removed() {
				to_remove.push(ship_object.collider);
				continue;
			}

			// Short-circuit continue if this ship isn't in game data
			// (which means it's playing a collapse sequence)
			if res.gx.get_ship(*handle).is_none() {
				let ship_object = self.ships.get_mut(handle).unwrap();
				ship_object.step(
					res,
					&mut self.rigid_body_set[ship_object.rigid_body],
					&mut self.collider_set[ship_object.collider],
				);
				continue;
			}

			// Compute new controls
			let controls;
			let b = self.ship_behaviors.get_mut(handle).unwrap();
			controls = b.update_controls(
				&res,
				&self.rigid_body_set,
				&self.ships,
				ship_object.rigid_body,
				*handle,
			);

			// Now re-borrow mutably to apply changes
			let ship_object = self.ships.get_mut(handle).unwrap();
			if let Some(controls) = controls {
				ship_object.controls = controls;
			}
			ship_object.step(
				res,
				&mut self.rigid_body_set[ship_object.rigid_body],
				&mut self.collider_set[ship_object.collider],
			);

			// If we're firing, try to fire each gun
			if ship_object.controls.guns {
				let ship_data = res.gx.get_ship_mut(ship_object.data_handle).unwrap();

				// TODO: don't allocate here. This is a hack to satisfy the borrow checker,
				// convert this to a refcell or do the replace dance.
				let pairs: Vec<(GunPoint, Option<OutfitHandle>)> = ship_data
					.get_outfits()
					.iter_gun_points()
					.map(|(p, o)| (p.clone(), o.clone()))
					.collect();

				for (gun, outfit) in pairs {
					if ship_data.fire_gun(res.ct, &gun) {
						projectiles.push((
							ship_object.data_handle,
							ship_object.rigid_body,
							gun.clone(),
							outfit.unwrap(),
						));
					}
				}
			}
		}

		// Remove ships that don't exist
		for c in to_remove {
			self.remove_ship(res, c);
		}

		// Create projectiles
		for (ship_dat, rigid_body, gun_point, outfit) in projectiles {
			let mut rng = rand::thread_rng();
			let rigid_body = self.get_rigid_body(rigid_body).unwrap();

			let ship_dat = res.gx.get_ship(ship_dat).unwrap();
			let ship_pos = util::rigidbody_position(rigid_body);
			let ship_rot = util::rigidbody_rotation(rigid_body);
			let ship_vel = util::rigidbody_velocity(rigid_body);
			let ship_ang = ship_rot.angle(Vector2 { x: 0.0, y: 1.0 });

			let pos = ship_pos + (Matrix2::from_angle(-ship_ang) * gun_point.pos.to_vec());

			let outfit = res.ct.get_outfit(outfit);
			let outfit = outfit.gun.as_ref().unwrap();

			let spread: Rad<f32> =
				Deg(rng.gen_range(-outfit.projectile.angle_rng.0..=outfit.projectile.angle_rng.0))
					.into();

			let vel = ship_vel
				+ (Matrix2::from_angle(-ship_ang + spread)
					* Vector2 {
						x: 0.0,
						y: outfit.projectile.speed
							+ rng.gen_range(
								-outfit.projectile.speed_rng..=outfit.projectile.speed_rng,
							),
					});

			let rigid_body = RigidBodyBuilder::kinematic_velocity_based()
				.translation(vector![pos.x, pos.y])
				.rotation(-ship_ang.0)
				.linvel(vector![vel.x, vel.y])
				.build();

			let collider = match &outfit.projectile.collider {
				ProjectileCollider::Ball(b) => ColliderBuilder::ball(b.radius)
					.sensor(true)
					.active_events(ActiveEvents::COLLISION_EVENTS)
					.build(),
			};

			let rigid_body = self.rigid_body_set.insert(rigid_body);
			let collider = self.collider_set.insert_with_parent(
				collider,
				rigid_body,
				&mut self.rigid_body_set,
			);

			self.projectiles.insert(
				collider.clone(),
				SySimProjectile::new(
					outfit.projectile.clone(),
					rigid_body,
					ship_dat.get_faction(),
					collider,
				),
			);
		}
	}

	/// Step this physics system by `t` seconds
	pub fn step(&mut self, mut res: StepResources) {
		res.timing.start_physics_ships();
		self.step_ships(&mut res);
		res.timing.mark_physics_ships();

		res.timing.start_physics_sim();

		// Update physics
		res.wrapper
			.step(res.t, &mut self.rigid_body_set, &mut self.collider_set);

		// Handle collision events
		while let Ok(event) = &res.wrapper.collision_queue.try_recv() {
			if event.started() {
				let a = event.collider1();
				let b = event.collider2();

				// If projectiles are part of this collision, make sure
				// `a` is one of them.
				let (a, b) = if self.projectiles.contains_key(&b) {
					(b, a)
				} else {
					(a, b)
				};

				let p = self.projectiles.get(&a);
				let s = self.ships.get_mut(match self.collider_ship_table.get(&b) {
					// If none, this ship is dead.
					Some(x) => x,
					None => continue,
				});
				if p.is_none() || s.is_none() {
					continue;
				}
				self.collide_projectile_ship(&mut res, a, b);
			}
		}

		// Delete projectiles
		let mut to_remove = Vec::new();
		for (c, p) in &mut self.projectiles {
			p.tick(res.t);
			if p.is_expired() {
				to_remove.push(*c);
			}
		}

		let mut rng = rand::thread_rng();
		for c in to_remove {
			let (pr, p) = self.remove_projectile(&mut res, c).unwrap();

			match &p.content.expire_effect {
				None => {}
				Some(x) => {
					let x = res.ct.get_effect(*x);
					let pos = util::rigidbody_position(&pr);
					let vel = util::rigidbody_velocity(&pr);
					let angle = util::rigidbody_rotation(&pr).angle(Vector2 { x: 1.0, y: 0.0 });

					let velocity = {
						let a = rng
							.gen_range(-x.velocity_scale_parent_rng..=x.velocity_scale_parent_rng);

						let velocity = (x.velocity_scale_parent + a) * vel;

						velocity
					};

					res.particles.push(ParticleBuilder::from_content(
						x,
						pos,
						-angle,
						velocity,
						Vector2::zero(),
					));
				}
			};
		}

		res.timing.mark_physics_sim();
	}
}

// Public getters
impl SystemSim {
	/// Get a ship physics object
	pub fn get_ship(&self, ship: GxShipHandle) -> Option<&SySimShip> {
		self.ships.get(&ship)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body(&self, r: RigidBodyHandle) -> Option<&RigidBody> {
		self.rigid_body_set.get(r)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body_mut(&mut self, r: RigidBodyHandle) -> Option<&mut RigidBody> {
		self.rigid_body_set.get_mut(r)
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ship_body(&self) -> impl Iterator<Item = (&SySimShip, &RigidBody)> + '_ {
		self.ships
			.values()
			.map(|x| (x, self.rigid_body_set.get(x.rigid_body).unwrap()))
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ships(&self) -> impl Iterator<Item = &SySimShip> + '_ {
		self.ships.values()
	}

	/// Iterate over all ships in this physics system
	pub fn iter_projectiles(&self) -> impl Iterator<Item = &SySimProjectile> + '_ {
		self.projectiles.values()
	}
}
